/* 
 * flatten.c --
 *
 *	This file contains the C-APIs for flattening a list.
 *
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: flatten.c,v 1.1 2000/01/14 23:44:16 aku Exp $
 */

#include "listxInt.h"

/*
 *----------------------------------------------------------------------
 *
 * Listx_Flatten --
 *
 *	Scans the incoming list and transforms all objects in it. The
 *	transformation result is placed into another list, which is
 *	then returned as result of the call.
 *	
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	See above.
 *
 *----------------------------------------------------------------------
 */

int
Listx_Flatten (interp, listObj, full, resultObj)
     Tcl_Interp*    interp;
     Tcl_Obj*       listObj;
     int            full;
     Tcl_Obj**      resultObj;
{
    int result;

    if (full) {
        /* Flatten the list as long as there are sublists to remove.
	 * The result will be a list without any substructure.
	 */

        Tcl_Obj* res;
	Tcl_Obj* in    = listObj;
	int      inLen, listLen;

	result = Tcl_ListObjLength(interp, in, &inLen);
	if (result != TCL_OK) {
	    return result;
	}

	while (1) {
	    Tcl_IncrRefCount (in);
	    result = Listx_Flatten (interp, in, 0, &res);
	    Tcl_DecrRefCount (in);

	    if (result != TCL_OK) {
	        return result;
	    }

	    /* If the result is of the same length as the
	     * input no structure was removed and we are
	     * done. Else we have to try to remove another
	     * level of structure.
	     */

	    result = Tcl_ListObjLength(interp, res, &listLen);

	    if (result != TCL_OK) {
	        return result;
	    }

	    if (listLen == inLen) {
	        *resultObj = res;
	        break;
	    }

	    in    = res;
	    inLen = listLen;
	}
    } else {
        /*
	 * Make sure the list argument is a list object and get its length and
	 * a pointer to its array of element pointers.
	 */

        Tcl_Obj*  flat;
	Tcl_Obj** listv;
	int       listc;
	long int  i;

	result = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);

	if (result != TCL_OK) {
	    return result;
	}

	/* Initialize the flattened list
	 */

	flat = Tcl_NewListObj(0, NULL);

	if (flat == (Tcl_Obj*) NULL) {
	    return TCL_ERROR;
	}
  
	/* Traverse the list. Try first to append an item as list, or rather its
	 * elements. If that one fails, append item as single element.
	 */

	for (i = 0; i < listc; i ++) {
	    if (Tcl_ListObjAppendList (interp, flat, listv [i]) == TCL_OK)
	        continue;

	    if (Tcl_ListObjAppendElement (interp, flat, listv [i]) == TCL_OK)
	        continue;

	    /* Discard the partial result.
	     */

	    Tcl_DecrRefCount (flat);
	    return TCL_ERROR;
	}

	*resultObj = flat;
    }

    return TCL_OK;
}



