/* 
 * index.c --
 *
 *	This file contains the C-APIs for selecting indexed items from a list.
 *
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: index.c,v 1.2 2000/01/14 23:44:16 aku Exp $
 */

#include "listxInt.h"

/*
 *----------------------------------------------------------------------
 *
 * Listx_SelectSFromList --
 *
 *	Selects the elements whose indices are specified by the selector
 *	structure from the list object and packs them into a new
 *	list object (of refcount 0).
 *
 * Results:
 *	A tcl standard result
 *
 * Side effects:
 *	resultObj is set to a new list object containing the selected
 *	elements. The value is not changed in case of problems.
 *
 *----------------------------------------------------------------------
 */

int
Listx_SelectSFromList (interp, listObj, ls, resultObj)
     Tcl_Interp*         interp;
     Tcl_Obj*            listObj;
     listSelector* CONST ls;
     Tcl_Obj**           resultObj;
{
  Tcl_Obj*  res, *item;
  int       listc, result, i, index, start, end, stride;
  Tcl_Obj **listv;

  /*
   * Make sure the list argument is a list object and get its length and
   * a pointer to its array of element pointers.
   */

  result = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);

  if (result != TCL_OK) {
    return result;
  }

  switch (ls->type) {
    case LIST_SEL_SINGLE:
      /* Return the single selected object. The caller has to bump the
       * reference counts.
       */

      result = ListxConvertToFullIndex (&ls->info.index, listc, &index);

      if (result == TCL_OK) {
	res = listv [index];
      } else {
	res = Tcl_NewObj ();
      }
      break;

    case LIST_SEL_MULTIPLE:
      /* Create a list containing all the selected items.
       */

      res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

      for (i=0; i < ls->info.multiple.num; i++) {
	result = ListxConvertToFullIndex (&ls->info.multiple.index [i], listc,
					  &index);

	if (result == TCL_OK) {
	  item = listv [index];
	} else {
	  item = Tcl_NewObj ();
	}

	result = Tcl_ListObjAppendElement (interp, res, item);

	if (result != TCL_OK) {
	  Tcl_DecrRefCount (res);
	  return TCL_ERROR;
	}
      }

      break;

    case LIST_SEL_RANGE:
      /* Create a list containing all the selected items.
       */

      ListxConvertToFullIndex (&ls->info.range.start, listc, &start);
      ListxConvertToFullIndex (&ls->info.range.end,   listc, &end);
      stride = ls->info.range.stride;

      if ((start > listc) || (end < 0) || (stride > listc)) {
	/* The range is empty. Return an empty object.
	 */

	res = Tcl_NewObj ();
      } else {
	/* Corrections in case of out-of-range indices.
	 */

	if (start < 0) {
	  start = (stride - ((-start) % stride)) % stride;
	}
	if (end >= listc) {
	  end = listc-1;
	}

	res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

	for (i=start; i <= end; i += stride) {
	  result = Tcl_ListObjAppendElement (interp, res, listv [i]);

	  if (result != TCL_OK) {
	    Tcl_DecrRefCount (res);
	    return TCL_ERROR;
	  }
	}
      }

      break;
  }

  *resultObj = res;
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Listx_SelectSFromListByObj --
 *
 *	Selects the elements whose indices are specified by the selector
 *	object from the list object and packs them into a new
 *	list object (of refcount 0).
 *
 * Results:
 *	A tcl standard result
 *
 * Side effects:
 *	resultObj is set to a new list object containing the selected
 *	elements. The value is not changed in case of problems.
 *
 *----------------------------------------------------------------------
 */

int
Listx_SelectSFromListByObj (interp, listObj, selectorObj, resultObj)
     Tcl_Interp* interp;
     Tcl_Obj*    listObj;
     Tcl_Obj*    selectorObj;
     Tcl_Obj**   resultObj;
{
  int           result, idx;
  listSelector* ls, ls_static;

  /* Special case code to avoid shimmering of 'int' objects.
   * Future: Use the 'Integer interface' if it is provided.
   *
   * In the special case a separate 'static' list selector is filled
   * with the relevant information and then given to the low level
   * code.
   */

  if (ListxIntType == selectorObj->typePtr) {

    ls = &ls_static;

    ls->type = LIST_SEL_SINGLE;
    result   = Tcl_GetIntFromObj (interp, selectorObj, &idx);

    if (result == TCL_OK) {
      ls->info.index.offsetFrom = LOFF_START;
      ls->info.index.location   = idx;
    }
  } else {
    result = Listx_GetSelectorFromObj (interp, selectorObj, &ls);
  }

  if (result != TCL_OK) {
    return result;
  }

  return Listx_SelectSFromList (interp, listObj, ls, resultObj);
}

/*
 *----------------------------------------------------------------------
 *
 * Listx_SelectMFromList --
 *
 *	Takes for each of the specified selector structures the elements
 *	selected by it, packs them into a new list and then makes these
 *	lists item in the final result, again a new list.
 *
 * Results:
 *	A standard tcl result code
 *
 * Side effects:
 *	resultObj is set to a new list object containing the lists
 *	containing the selected elements. The value is not changed if
 *	there were problems.
 *
 *----------------------------------------------------------------------
 */

int
Listx_SelectMFromList (interp, listObj, lsc, lsv, resultObj)
     Tcl_Interp*          interp;
     Tcl_Obj*             listObj;
     int                  lsc;
     listSelector** CONST lsv;
     Tcl_Obj**            resultObj;
{
  int result, i;
  Tcl_Obj* res;
  Tcl_Obj* item;

  res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

  for (i=0; i < lsc; i++) {
    result = Listx_SelectSFromList (interp, listObj, lsv [i], &item);

    if (result != TCL_OK) {
      Tcl_DecrRefCount (res);
      return TCL_ERROR;
    }

    result = Tcl_ListObjAppendElement (interp, res, item);

    if (result != TCL_OK) {
      Tcl_DecrRefCount (res);
      return TCL_ERROR;
    }
  }

  *resultObj = res;
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Listx_SelectMFromListByObj --
 *
 *	Takes for each of the specified selector objects the elements
 *	selected by it, packs them into a new list and then makes these
 *	lists item in the final result, again a new list.
 *
 * Results:
 *	A standard tcl result code
 *
 * Side effects:
 *	resultObj is set to a new list object containing the lists
 *	containing the selected elements. The value is not changed if
 *	there were problems.
 *
 *----------------------------------------------------------------------
 */

int
Listx_SelectMFromListByObj (interp, listObj, soc, sov, resultObj)
     Tcl_Interp* interp;
     Tcl_Obj*    listObj;
     int         soc;
     Tcl_Obj**   sov;
     Tcl_Obj**   resultObj;
{
  int            result, i, idx;
  listSelector** ls, *ls_int;

  ls     = (listSelector**) Tcl_Alloc (soc * sizeof (listSelector*));
  ls_int = (listSelector*)  Tcl_Alloc (soc * sizeof (listSelector));

  for (i=0; i< soc; i++) {

    /* Special case code to avoid shimmering of 'int' objects.
     * Future: Use the 'Integer interface' if it is provided.
     * See [x] too.
     */

    if (ListxIntType == sov [i]->typePtr) {

      ls [i] = &ls_int [i];
      ls [i]->type = LIST_SEL_SINGLE;

      result = Tcl_GetIntFromObj (interp, sov [i], &idx);

      if (result == TCL_OK) {
	ls [i]->info.index.offsetFrom = LOFF_START;
	ls [i]->info.index.location   = idx;
      }
    } else {
      result = Listx_GetSelectorFromObj (interp, sov [i], &(ls [i]));
    }

    if (result != TCL_OK) {
      Tcl_Free ((char*) ls);
      Tcl_Free ((char*) ls_int);
      return result;
    }
  }

  result = Listx_SelectMFromList (interp, listObj, soc, ls, resultObj);

  Tcl_Free ((char*) ls);
  Tcl_Free ((char*) ls_int);
  return result;
}

