/* 
 * listx.c --
 *
 *	This file contains the top-level command routine for the
 *	'::tcl::listx' command.
 *
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: listx.c,v 1.2 2000/01/14 23:44:16 aku Exp $
 */

#include "listxInt.h"
#ifndef USE_EVALOBJV
#define USE_EVALOBJV 0
#endif

/*
 * During execution of the "sort" command, structures of the following
 * type are used to arrange the objects being sorted into a collection
 * of linked lists.
 */

typedef struct SortElement {
    Tcl_Obj *objPtr;			/* Object being sorted. */
    int count;				/* number of same elements in list */
    struct SortElement *nextPtr;        /* Next element in the list, or
					 * NULL for end of list. */
} SortElement;

/*
 * The "sort" command needs to pass certain information down to the
 * function that compares two list elements, and the comparison function
 * needs to pass success or failure information back up to the top-level
 * "lsort" command.  The following structure is used to pass this
 * information.
 */

typedef struct SortInfo {
    int isIncreasing;		/* Nonzero means sort in increasing order. */
    int sortMode;		/* The sort mode.  One of SORTMODE_*
				 * values defined below */
    Tcl_Obj *compareCmdPtr;     /* The Tcl comparison command when sortMode
				 * is SORTMODE_COMMAND.  Pre-initialized to
				 * hold base of command.*/
    int index;			/* If the -index option was specified, this
				 * holds the index of the list element
				 * to extract for comparison.  If -index
				 * wasn't specified, this is -1. */
    Tcl_Interp *interp;		/* The interpreter in which the sortis
				 * being done. */
    int resultCode;		/* Completion code for the lsort command.
				 * If an error occurs during the sort this
				 * is changed from TCL_OK to  TCL_ERROR. */
} SortInfo;

/*
 * The "sortMode" field of the SortInfo structure can take on any of the
 * following values.
 */

#define SORTMODE_ASCII      0
#define SORTMODE_INTEGER    1
#define SORTMODE_REAL       2
#define SORTMODE_COMMAND    3
#define SORTMODE_DICTIONARY 4


/*
 * Prototypes for local procedures defined in this file:
 */

static int		DictionaryCompare _ANSI_ARGS_((char *left,
			    char *right));
static SortElement *    MergeSort _ANSI_ARGS_((SortElement *headPt,
			    SortInfo *infoPtr));
static SortElement *    MergeLists _ANSI_ARGS_((SortElement *leftPtr,
			    SortElement *rightPtr, SortInfo *infoPtr));
static int		SortCompare _ANSI_ARGS_((Tcl_Obj *firstPtr,
			    Tcl_Obj *second, SortInfo *infoPtr));

static int              ListxScriptMapProc _ANSI_ARGS_ ((ClientData clientData,
	 Tcl_Interp* interp, Tcl_Obj* inObj, Tcl_Obj** resultObj));

static int             ListxScriptFoldProc _ANSI_ARGS_ ((ClientData clientData,
	 Tcl_Interp* interp, Tcl_Obj* accumulatorObj, Tcl_Obj* inObj,
	 Tcl_Obj** resultObj));


static int	AppendObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	AssignObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	ConcatObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	CreateObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	EmptyObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	FlattenObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	FoldObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	IndexObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	InsertObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	JoinObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	LengthObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	MapObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	MergeObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	RangeObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	RemoveObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	ReplaceObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	ReverseObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	SearchObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	SortObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	SplitObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	UniqueObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
static int	UnknownObjCmd _ANSI_ARGS_((ClientData dummy,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));

/*
 * Global variable holding a reference to the 'ListObject'-Type.
 * Initialized in --> init.c:Listx_Init ().
 */

Tcl_ObjType* ListxListType      = (Tcl_ObjType*) NULL;
Tcl_Obj*     ListxEmptyObj      = (Tcl_Obj*)     NULL;
Tcl_Obj*     ListxUnknownCmdObj = (Tcl_Obj*)     NULL;
Tcl_ObjType* ListxIntType       = (Tcl_ObjType*) NULL;

#define tclEmptyStringRep (ListxEmptyObj->bytes)

/*
 *----------------------------------------------------------------------
 *
 * ListxListxObjCmd --
 *
 *	This command provides the extended list of list commands.
 *	It implements the ::listx command. See the user documentation
 *	for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

extern int
ListxListxObjCmd (dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    int index, res;

    static char *optionStrings[] = {
        "append",  "assign",  "concat", "create", "empty",
	"flatten", "fold",    "index",  "insert", "join",
	"length",  "map",     "merge",  "range",  "remove",
	"replace", "reverse", "search", "sort",   "split",
	"unique",
	NULL
    };
    enum options {
        LIST_APPEND,  LIST_ASSIGN,  LIST_CONCAT,  LIST_CREATE,  LIST_EMPTY,
	LIST_FLATTEN, LIST_FOLD,    LIST_INDEX,   LIST_INSERT,  LIST_JOIN,
	LIST_LENGTH,  LIST_MAP,     LIST_MERGE,   LIST_RANGE,   LIST_REMOVE,
	LIST_REPLACE, LIST_REVERSE, LIST_SEARCH,  LIST_SORT,    LIST_SPLIT,
	LIST_UNIQUE
    };

    if (objc < 2) {
    	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
        return TCL_ERROR;
    }
    if (Tcl_GetIndexFromObj(interp, objv[1], optionStrings, "option", 0,
	    &index) != TCL_OK) {
        return UnknownObjCmd (dummy, interp, objc, objv);
    }

    switch ((enum options) index) {
        case LIST_APPEND:
	    res = AppendObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_ASSIGN:
	    res = AssignObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_CONCAT:
	    res = ConcatObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_CREATE:
	    res = CreateObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_EMPTY:
	    res = EmptyObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_FLATTEN:
	    res = FlattenObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_FOLD:
	    res = FoldObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_INDEX:
	    res = IndexObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_INSERT:
	    res = InsertObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_JOIN:
	    res = JoinObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_LENGTH:
	    res = LengthObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_MAP:
	    res = MapObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_MERGE:
	    res = MergeObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_RANGE:
	    res = RangeObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_REMOVE:
	    res = RemoveObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_REPLACE:
	    res = ReplaceObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_REVERSE:
	    res = ReverseObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_SEARCH:
	    res = SearchObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_SORT:
	    res = SortObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_SPLIT:
	    res = SplitObjCmd (dummy, interp, objc-1, objv+1);
	    break;
        case LIST_UNIQUE:
	    res = UniqueObjCmd (dummy, interp, objc-1, objv+1);
	    break;
    }

    return res;
}

/*
 *----------------------------------------------------------------------
 *
 * JoinObjCmd --
 *
 *	This procedure is invoked to process the "join" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
JoinObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* The argument objects. */
{
    char *joinString, *bytes;
    int joinLength, listLen, length, i, result;
    Tcl_Obj **elemPtrs;
    Tcl_Obj *resObjPtr;

    if (objc == 2) {
	joinString = " ";
	joinLength = 1;
    } else if (objc == 3) {
	joinString = Tcl_GetStringFromObj(objv[2], &joinLength);
    } else {
	Tcl_WrongNumArgs(interp, 2, objv-1, "list ?joinString?");
	return TCL_ERROR;
    }

    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    result = Tcl_ListObjGetElements(interp, objv[1], &listLen, &elemPtrs);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * Now concatenate strings to form the "joined" result. We append
     * directly into the interpreter's result object.
     */

    resObjPtr = Tcl_GetObjResult(interp);

    for (i = 0;  i < listLen;  i++) {
	bytes = Tcl_GetStringFromObj(elemPtrs[i], &length);
	if (i > 0) {
	    Tcl_AppendToObj(resObjPtr, joinString, joinLength);
	}
	Tcl_AppendToObj(resObjPtr, bytes, length);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * IndexObjCmd --
 *
 *	This procedure is invoked to process the "index" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
static int
IndexObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj *listPtr, *res;
    Tcl_Obj **listv;
    int listc, result;

    /*
     * Recognized syntax:	list ?selector?...
     *
     * And a selector is either an index or a list of indices.
     * An index is a number or an expression of the form end?-number?.
     */

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "list ?selector?...");
	return TCL_ERROR;
    }

    /*
     * Convert the first argument to a list if necessary.
     */

    listPtr = objv[1];
    result = Tcl_ListObjGetElements(interp, listPtr, &listc, &listv);

    if (result != TCL_OK) {
	return result;
    }

    if (objc == 2) {
      /* No selecting objects, just the list. The result is empty.
       */

      Tcl_SetObjResult (interp, Tcl_NewObj ());
      return TCL_OK;

    } else if (objc == 3) {
      /* A single selecting object. The result is specified by that
       * object.
       */

      result = Listx_SelectSFromListByObj (interp, listPtr, objv [2], &res);

    } else {
      /* More than one selector. Take all selections and make them
       * the final result.
       */

      result = Listx_SelectMFromListByObj (interp, listPtr,
					   objc-2, (Tcl_Obj**) objv+2,
					   &res);
    }

    if (result != TCL_OK) {
      return result;
    }

    /*
     * Set the interpreter's object result to the selection
     */

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * InsertObjCmd --
 *
 *	This procedure is invoked to process the "insert" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A new Tcl list object formed by inserting zero or more elements 
 *	into a list.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
InsertObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    register int objc;		/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj *listPtr, *resultPtr;
    Tcl_ObjType *typePtr;
    int index, isDuplicate, len, result;
   
    if (objc < 4) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "list index element ?element ...?");
	return TCL_ERROR;
    }

    /*
     * Get the index first since, if a conversion to int is needed, it
     * will invalidate the list's internal representation.
     */

    result = Tcl_ListObjLength(interp, objv[1], &len);
    if (result != TCL_OK) {
	return result;
    }

    result = TclGetIntForIndex(interp, objv[2], /*endValue*/ len, &index);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * If the list object is unshared we can modify it directly. Otherwise
     * we create a copy to modify: this is "copy on write". We create the
     * duplicate directly in the interpreter's object result.
     */
    
    listPtr = objv[1];
    isDuplicate = 0;
    if (Tcl_IsShared(listPtr)) {
	/*
	 * The following code must reflect the logic in Tcl_DuplicateObj()
	 * except that it must duplicate the list object directly into the
	 * interpreter's result.
	 */
	
	Tcl_ResetResult(interp);
	resultPtr = Tcl_GetObjResult(interp);
	typePtr = listPtr->typePtr;
	if (listPtr->bytes == NULL) {
	    resultPtr->bytes = NULL;
	} else if (listPtr->bytes != tclEmptyStringRep) {
	    len = listPtr->length;
	    TclInitStringRep(resultPtr, listPtr->bytes, len);
	}
	if (typePtr != NULL) {
	    if (typePtr->dupIntRepProc == NULL) {
		resultPtr->internalRep = listPtr->internalRep;
		resultPtr->typePtr = typePtr;
	    } else {
		(*typePtr->dupIntRepProc)(listPtr, resultPtr);
	    }
	}
	listPtr = resultPtr;
	isDuplicate = 1;
    }
    
    if ((objc == 4) && (index == INT_MAX)) {
	/*
	 * Special case: insert one element at the end of the list.
	 */

	result = Tcl_ListObjAppendElement(interp, listPtr, objv[3]);
    } else if (objc > 3) {
	result = Tcl_ListObjReplace(interp, listPtr, index, 0,
				    (objc-3), &(objv[3]));
    }
    if (result != TCL_OK) {
	return result;
    }
    
    /*
     * Set the interpreter's object result.
     */

    if (!isDuplicate) {
	Tcl_SetObjResult(interp, listPtr);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateObjCmd --
 *
 *	This procedure is invoked to process the "create" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
CreateObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    register int objc;			/* Number of arguments. */
    register Tcl_Obj *CONST objv[];	/* The argument objects. */
{
    /*
     * If there are no list elements, the result is an empty object.
     * Otherwise modify the interpreter's result object to be a list object.
     */
    
    if (objc > 1) {
	Tcl_SetListObj(Tcl_GetObjResult(interp), (objc-1), &(objv[1]));
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * LengthObjCmd --
 *
 *	This procedure is invoked to process the "length" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
LengthObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    register Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    int listLen, result;

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "list");
	return TCL_ERROR;
    }

    result = Tcl_ListObjLength(interp, objv[1], &listLen);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * Set the interpreter's object result to an integer object holding the
     * length. 
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), listLen);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * RangeObjCmd --
 *
 *	This procedure is invoked to process the "range" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
RangeObjCmd(notUsed, interp, objc, objv)
    ClientData notUsed;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    register Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj *listPtr;
    Tcl_Obj **elemPtrs;
    int listLen, first, last, numElems, result;

    if (objc != 4) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "list first last");
	return TCL_ERROR;
    }

    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    listPtr = objv[1];
    result = Tcl_ListObjGetElements(interp, listPtr, &listLen, &elemPtrs);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * Get the first and last indexes.
     */

    result = TclGetIntForIndex(interp, objv[2], /*endValue*/ (listLen - 1),
	    &first);
    if (result != TCL_OK) {
	return result;
    }
    if (first < 0) {
	first = 0;
    }

    result = TclGetIntForIndex(interp, objv[3], /*endValue*/ (listLen - 1),
	    &last);
    if (result != TCL_OK) {
	return result;
    }
    if (last >= listLen) {
	last = (listLen - 1);
    }
    
    if (first > last) {
	return TCL_OK;		/* the result is an empty object */
    }

    /*
     * Make sure listPtr still refers to a list object. It might have been
     * converted to an int above if the argument objects were shared.
     */  

    if (listPtr->typePtr != ListxListType) {
        result = Tcl_ListObjGetElements(interp, listPtr, &listLen,
                &elemPtrs);
        if (result != TCL_OK) {
            return result;
        }
    }

    /*
     * Extract a range of fields. We modify the interpreter's result object
     * to be a list object containing the specified elements.
     */

    numElems = (last - first + 1);
    Tcl_SetListObj(Tcl_GetObjResult(interp), numElems, &(elemPtrs[first]));
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ReplaceObjCmd --
 *
 *	This procedure is invoked to process the "replace" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A new Tcl list object formed by replacing zero or more elements of
 *	a list.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
ReplaceObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    register Tcl_Obj *listPtr;
    int createdNewObj, first, last, listLen, numToDelete;
    int firstArgLen, result;
    char *firstArg;

    if (objc < 4) {
	Tcl_WrongNumArgs(interp, 2, objv-1,
		"list first last ?element element ...?");
	return TCL_ERROR;
    }

    /*
     * If the list object is unshared we can modify it directly, otherwise
     * we create a copy to modify: this is "copy on write".
     */
    
    listPtr = objv[1];
    createdNewObj = 0;
    if (Tcl_IsShared(listPtr)) {
	listPtr = Tcl_DuplicateObj(listPtr);
	createdNewObj = 1;
    }
    result = Tcl_ListObjLength(interp, listPtr, &listLen);
    if (result != TCL_OK) {
        errorReturn:
	if (createdNewObj) {
	    Tcl_DecrRefCount(listPtr); /* free unneeded obj */
	}
	return result;
    }

    /*
     * Get the first and last indexes.
     */

    result = TclGetIntForIndex(interp, objv[2], /*endValue*/ (listLen - 1),
	    &first);
    if (result != TCL_OK) {
	goto errorReturn;
    }
    firstArg = Tcl_GetStringFromObj(objv[2], &firstArgLen);

    result = TclGetIntForIndex(interp, objv[3], /*endValue*/ (listLen - 1),
	    &last);
    if (result != TCL_OK) {
	goto errorReturn;
    }

    if (first < 0)  {
    	first = 0;
    }
    if ((first >= listLen) && (listLen > 0)
	    && (strncmp(firstArg, "end", (unsigned) firstArgLen) != 0)) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"list doesn't contain element ",
		Tcl_GetString(objv[2]), (int *) NULL);
	result = TCL_ERROR;
	goto errorReturn;
    }
    if (last >= listLen) {
    	last = (listLen - 1);
    }
    if (first <= last) {
	numToDelete = (last - first + 1);
    } else {
	numToDelete = 0;
    }

    if (objc > 4) {
	result = Tcl_ListObjReplace(interp, listPtr, first, numToDelete,
	        (objc-4), &(objv[4]));
    } else {
	result = Tcl_ListObjReplace(interp, listPtr, first, numToDelete,
		0, NULL);
    }
    if (result != TCL_OK) {
	goto errorReturn;
    }

    /*
     * Set the interpreter's object result. 
     */

    Tcl_SetObjResult(interp, listPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SearchObjCmd --
 *
 *	This procedure is invoked to process the "search" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
SearchObjCmd(dummy, interp, objc, objv)
    ClientData dummy;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument values. */
{
    int          result, listc, nocase, inverted, all, i;
    lSearchModes mode;
    listSelector ls;
    Tcl_Obj*     patObj, *listObj, **listv;

    static char *options [] = {
        "-all", "-exact",  "-glob",  "-nocase",
	"-not", "-regexp", "--",
	NULL
    };
    enum options {
        SEARCH_ALL, SEARCH_EXACT,  SEARCH_GLOB,  SEARCH_NOCASE,
	SEARCH_NOT, SEARCH_REGEXP, SEARCH_LAST
    };

    if (objc < 3) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "?options? list pattern...");
	return TCL_ERROR;
    }

    all      = 0;
    mode     = LSEARCH_GLOB;
    inverted = 0;
    nocase   = 0;

    for (i = 1; i < objc; i++) {
	char* name;
	int   index;

	name = Tcl_GetString(objv[i]);
	if (name[0] != '-') {
	    break;
	}
	if (Tcl_GetIndexFromObj(interp, objv[i], options, "option", TCL_EXACT,
		&index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch ((enum options) index) {
            case SEARCH_ALL: {
	        all = 1;
		break;
	    }
	    case SEARCH_EXACT: {
	        mode = LSEARCH_EXACT;
		break;
	    }
	    case SEARCH_GLOB: {
	        mode = LSEARCH_GLOB;
		break;
	    }
	    case SEARCH_NOCASE: {
	        nocase = 1;
		break;
	    }
	    case SEARCH_NOT: {
	        inverted = 1;
		break;
	    }
	    case SEARCH_REGEXP: {
	        mode = LSEARCH_REGEXP;
		break;
	    }
	    case SEARCH_LAST: {
		i++;
		goto endOfForLoop;
	    }
	}
    }

    endOfForLoop:
    if (objc - i < 2) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "?options? list pattern...");
	return TCL_ERROR;
    }

    objc -= i;
    objv += i;

    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    listObj = objv [0];
    result  = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);

    if (result != TCL_OK) {
	return result;
    }

    /*
     * Single or multiple patterns ?
     * All or first index ?
     */

    if (objc == 2) {
        /* List and a single pattern */

      patObj = objv [1];

      result = Listx_Search (interp, listObj, patObj, 0, mode, inverted,
			     nocase, all, &ls);

      if (result != TCL_OK) {
	return result;
      }

      if (!all &&
	  (ls.type == LIST_SEL_MULTIPLE) &&
	  (ls.info.multiple.num == 0)) {
	Tcl_SetObjResult (interp, Tcl_NewIntObj (-1));
      } else {
	Tcl_SetObjResult (interp, Listx_NewSelectorObj (&ls));
      }
      return TCL_OK;
    } else {
        /* List and multiple patterns */

        Tcl_Obj* res, *item;

	res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);
	if (res == (Tcl_Obj*) NULL) {
	    return TCL_ERROR;
	}

        for (i=1; i < objc; i++) {
	    patObj = objv [i];

	    result = Listx_Search (interp, listObj, patObj, 0, mode, inverted,
				   nocase, all, &ls);

	    if (result != TCL_OK) {
	      Tcl_DecrRefCount (res);
	      return result;
	    }

	    if (!all &&
		(ls.type == LIST_SEL_MULTIPLE) &&
		(ls.info.multiple.num == 0)) {
	      item = Tcl_NewIntObj (-1);
	    } else {
	      item = Listx_NewSelectorObj (&ls);
	    }

	    result = Tcl_ListObjAppendElement(interp, res, item);

	    if (result != TCL_OK) {
	        return result;
	    }
	}

	Tcl_SetObjResult (interp, res);
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SortObjCmd --
 *
 *	This procedure is invoked to process the "sort" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
SortObjCmd(dummy, interp, objc, objv)
    ClientData dummy;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument values. */
{
    int i, index, unique;
    Tcl_Obj *resultPtr;
    int length;
    Tcl_Obj *cmdPtr, **listObjPtrs;
    SortElement *elementArray;
    SortElement *elementPtr;        
    SortInfo sortInfo;                  /* Information about this sort that
                                         * needs to be passed to the 
                                         * comparison function */
    static char *switches[] =
	    {"-ascii", "-command", "-decreasing", "-dictionary", "-increasing",
	     "-index", "-integer", "-real", "-unique", (char *) NULL};

    resultPtr = Tcl_GetObjResult(interp);
    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "?options? list");
	return TCL_ERROR;
    }

    /*
     * Parse arguments to set up the mode for the sort.
     */

    sortInfo.isIncreasing = 1;
    sortInfo.sortMode = SORTMODE_ASCII;
    sortInfo.index = -1;
    sortInfo.interp = interp;
    sortInfo.resultCode = TCL_OK;
    cmdPtr = NULL;
    unique = 0;

    for (i = 1; i < objc-1; i++) {
	if (Tcl_GetIndexFromObj(interp, objv[i], switches, "option", 0, &index)
		!= TCL_OK) {
	    return TCL_ERROR;
	}
	switch (index) {
	    case 0:			/* -ascii */
		sortInfo.sortMode = SORTMODE_ASCII;
		break;
	    case 1:			/* -command */
		if (i == (objc-2)) {
		    Tcl_AppendToObj(resultPtr,
			    "\"-command\" option must be followed by comparison command",
			    -1);
		    return TCL_ERROR;
		}
		sortInfo.sortMode = SORTMODE_COMMAND;
		cmdPtr = objv[i+1];
		i++;
		break;
	    case 2:			/* -decreasing */
		sortInfo.isIncreasing = 0;
		break;
	    case 3:			/* -dictionary */
		sortInfo.sortMode = SORTMODE_DICTIONARY;
		break;
	    case 4:			/* -increasing */
		sortInfo.isIncreasing = 1;
		break;
	    case 5:			/* -index */
		if (i == (objc-2)) {
		    Tcl_AppendToObj(resultPtr,
			    "\"-index\" option must be followed by list index",
			    -1);
		    return TCL_ERROR;
		}
		if (TclGetIntForIndex(interp, objv[i+1], -2, &sortInfo.index)
			!= TCL_OK) {
		    return TCL_ERROR;
		}
		cmdPtr = objv[i+1];
		i++;
		break;
	    case 6:			/* -integer */
		sortInfo.sortMode = SORTMODE_INTEGER;
		break;
	    case 7:			/* -real */
		sortInfo.sortMode = SORTMODE_REAL;
		break;
	    case 8:                     /* -unique */
	        unique = 1;
		break;
	}
    }
    if (sortInfo.sortMode == SORTMODE_COMMAND) {
	/*
	 * The existing command is a list. We want to flatten it, append
	 * two dummy arguments on the end, and replace these arguments
	 * later.
	 */

        Tcl_Obj *newCommandPtr = Tcl_DuplicateObj(cmdPtr);
	Tcl_Obj *newObjPtr     = Tcl_NewObj();

	Tcl_IncrRefCount(newCommandPtr);
	if (Tcl_ListObjAppendElement(interp, newCommandPtr, Tcl_NewObj()) 
	        != TCL_OK) {
	    Tcl_DecrRefCount(newCommandPtr);
	    Tcl_DecrRefCount(newObjPtr);
	    return TCL_ERROR;
	}
	Tcl_ListObjAppendElement(interp, newCommandPtr, Tcl_NewObj());
	sortInfo.compareCmdPtr = newCommandPtr;
    }

    sortInfo.resultCode = Tcl_ListObjGetElements(interp, objv[objc-1],
	    &length, &listObjPtrs);
    if (sortInfo.resultCode != TCL_OK) {
	goto done;
    }
    if (length <= 0) {
        return TCL_OK;
    }
    elementArray = (SortElement *) ckalloc(length * sizeof(SortElement));
    for (i=0; i < length; i++){
	elementArray[i].objPtr  = listObjPtrs[i];
	elementArray[i].count   = 0;
	elementArray[i].nextPtr = &elementArray[i+1];
    }
    elementArray[length-1].nextPtr = NULL;
    elementPtr = MergeSort(elementArray, &sortInfo);
    if (sortInfo.resultCode == TCL_OK) {
	/*
	 * Note: must clear the interpreter's result object: it could
	 * have been set by the -command script.
	 */

	Tcl_ResetResult(interp);
	resultPtr = Tcl_GetObjResult(interp);

	if (unique) {
	    for (; elementPtr != NULL; elementPtr = elementPtr->nextPtr){
	        if (elementPtr->count == 0) {
		    Tcl_ListObjAppendElement(interp, resultPtr,
			     elementPtr->objPtr);
		}
	    }
	} else {
	    for (; elementPtr != NULL; elementPtr = elementPtr->nextPtr){
	        Tcl_ListObjAppendElement(interp, resultPtr,
			 elementPtr->objPtr);
	    }
	}
    }
    ckfree((char*) elementArray);

    done:
    if (sortInfo.sortMode == SORTMODE_COMMAND) {
	Tcl_DecrRefCount(sortInfo.compareCmdPtr);
	sortInfo.compareCmdPtr = NULL;
    }
    return sortInfo.resultCode;
}

/*
 *----------------------------------------------------------------------
 *
 * MergeSort -
 *
 *	This procedure sorts a linked list of SortElement structures
 *	use the merge-sort algorithm.
 *
 * Results:
 *      A pointer to the head of the list after sorting is returned.
 *
 * Side effects:
 *	None, unless a user-defined comparison command does something
 *	weird.
 *
 *----------------------------------------------------------------------
 */

static SortElement *
MergeSort(headPtr, infoPtr)
    SortElement *headPtr;               /* First element on the list */
    SortInfo *infoPtr;                  /* Information needed by the
                                         * comparison operator */
{
    /*
     * The subList array below holds pointers to temporary lists built
     * during the merge sort.  Element i of the array holds a list of
     * length 2**i.
     */

#   define NUM_LISTS 30
    SortElement *subList[NUM_LISTS];
    SortElement *elementPtr;
    int i;

    for(i = 0; i < NUM_LISTS; i++){
        subList[i] = NULL;
    }
    while (headPtr != NULL) {
	elementPtr = headPtr;
	headPtr = headPtr->nextPtr;
	elementPtr->nextPtr = 0;
	for (i = 0; (i < NUM_LISTS) && (subList[i] != NULL); i++){
	    elementPtr = MergeLists(subList[i], elementPtr, infoPtr);
	    subList[i] = NULL;
	}
	if (i >= NUM_LISTS) {
	    i = NUM_LISTS-1;
	}
	subList[i] = elementPtr;
    }
    elementPtr = NULL;
    for (i = 0; i < NUM_LISTS; i++){
        elementPtr = MergeLists(subList[i], elementPtr, infoPtr);
    }
    return elementPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * MergeLists -
 *
 *	This procedure combines two sorted lists of SortElement structures
 *	into a single sorted list.
 *
 * Results:
 *      The unified list of SortElement structures.
 *
 * Side effects:
 *	None, unless a user-defined comparison command does something
 *	weird.
 *
 *----------------------------------------------------------------------
 */

static SortElement *
MergeLists(leftPtr, rightPtr, infoPtr)
    SortElement *leftPtr;               /* First list to be merged; may be
					 * NULL. */
    SortElement *rightPtr;              /* Second list to be merged; may be
					 * NULL. */
    SortInfo *infoPtr;                  /* Information needed by the
                                         * comparison operator. */
{
    SortElement *headPtr;
    SortElement *tailPtr;
    int          cmp;

    if (leftPtr == NULL) {
        return rightPtr;
    }
    if (rightPtr == NULL) {
        return leftPtr;
    }
    cmp = SortCompare(leftPtr->objPtr, rightPtr->objPtr, infoPtr);
    if (cmp > 0) {
	tailPtr = rightPtr;
	rightPtr = rightPtr->nextPtr;
    } else {
	if (cmp == 0) {
	    leftPtr->count++;
	}
	tailPtr = leftPtr;
	leftPtr = leftPtr->nextPtr;
    }
    headPtr = tailPtr;
    while ((leftPtr != NULL) && (rightPtr != NULL)) {
      cmp = SortCompare(leftPtr->objPtr, rightPtr->objPtr, infoPtr);
	if (cmp > 0) {
	    tailPtr->nextPtr = rightPtr;
	    tailPtr = rightPtr;
	    rightPtr = rightPtr->nextPtr;
	} else {
	    if (cmp == 0) {
		leftPtr->count++;
	    }
	    tailPtr->nextPtr = leftPtr;
	    tailPtr = leftPtr;
	    leftPtr = leftPtr->nextPtr;
	}
    }
    if (leftPtr != NULL) {
       tailPtr->nextPtr = leftPtr;
    } else {
       tailPtr->nextPtr = rightPtr;
    }
    return headPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * SortCompare --
 *
 *	This procedure is invoked by MergeLists to determine the proper
 *	ordering between two elements.
 *
 * Results:
 *      A negative results means the the first element comes before the
 *      second, and a positive results means that the second element
 *      should come first.  A result of zero means the two elements
 *      are equal and it doesn't matter which comes first.
 *
 * Side effects:
 *	None, unless a user-defined comparison command does something
 *	weird.
 *
 *----------------------------------------------------------------------
 */

static int
SortCompare(objPtr1, objPtr2, infoPtr)
    Tcl_Obj *objPtr1, *objPtr2;		/* Values to be compared. */
    SortInfo *infoPtr;                  /* Information passed from the
                                         * top-level "sort" command */
{
    int order, listLen, index;
    Tcl_Obj *objPtr;
    char buffer[TCL_INTEGER_SPACE];

    order = 0;
    if (infoPtr->resultCode != TCL_OK) {
	/*
	 * Once an error has occurred, skip any future comparisons
	 * so as to preserve the error message in sortInterp->result.
	 */

	return order;
    }
    if (infoPtr->index != -1) {
	/*
	 * The "-index" option was specified.  Treat each object as a
	 * list, extract the requested element from each list, and
	 * compare the elements, not the lists.  The special index "end"
	 * is signaled here with a large negative index.
	 */

	if (Tcl_ListObjLength(infoPtr->interp, objPtr1, &listLen) != TCL_OK) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (infoPtr->index < -1) {
	    index = listLen - 1;
	} else {
	    index = infoPtr->index;
	}

	if (Tcl_ListObjIndex(infoPtr->interp, objPtr1, index, &objPtr)
		!= TCL_OK) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (objPtr == NULL) {
	    objPtr = objPtr1;
	    missingElement:
	    TclFormatInt(buffer, infoPtr->index);
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(infoPtr->interp),
			"element ", buffer, " missing from sublist \"",
			Tcl_GetString(objPtr), "\"", (char *) NULL);
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	objPtr1 = objPtr;

	if (Tcl_ListObjLength(infoPtr->interp, objPtr2, &listLen) != TCL_OK) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (infoPtr->index < -1) {
	    index = listLen - 1;
	} else {
	    index = infoPtr->index;
	}

	if (Tcl_ListObjIndex(infoPtr->interp, objPtr2, index, &objPtr)
		!= TCL_OK) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (objPtr == NULL) {
	    objPtr = objPtr2;
	    goto missingElement;
	}
	objPtr2 = objPtr;
    }
    if (infoPtr->sortMode == SORTMODE_ASCII) {
	order = strcmp(Tcl_GetString(objPtr1), Tcl_GetString(objPtr2));
    } else if (infoPtr->sortMode == SORTMODE_DICTIONARY) {
	order = DictionaryCompare(
		Tcl_GetString(objPtr1),	Tcl_GetString(objPtr2));
    } else if (infoPtr->sortMode == SORTMODE_INTEGER) {
	int a, b;

	if ((Tcl_GetIntFromObj(infoPtr->interp, objPtr1, &a) != TCL_OK)
		|| (Tcl_GetIntFromObj(infoPtr->interp, objPtr2, &b)
		!= TCL_OK)) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (a > b) {
	    order = 1;
	} else if (b > a) {
	    order = -1;
	}
    } else if (infoPtr->sortMode == SORTMODE_REAL) {
	double a, b;

	if ((Tcl_GetDoubleFromObj(infoPtr->interp, objPtr1, &a) != TCL_OK)
	      || (Tcl_GetDoubleFromObj(infoPtr->interp, objPtr2, &b)
		      != TCL_OK)) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (a > b) {
	    order = 1;
	} else if (b > a) {
	    order = -1;
	}
    } else {
	Tcl_Obj **objv, *paramObjv[2];
	int objc;

	paramObjv[0] = objPtr1;
	paramObjv[1] = objPtr2;

  	/*
 	 * We made space in the command list for the two things to
	 * compare. Replace them and evaluate the result.
	 */

	Tcl_ListObjLength(infoPtr->interp, infoPtr->compareCmdPtr, &objc);
	Tcl_ListObjReplace(infoPtr->interp, infoPtr->compareCmdPtr, objc - 2,
		2, 2, paramObjv);
   	Tcl_ListObjGetElements(infoPtr->interp, infoPtr->compareCmdPtr,
		&objc, &objv);

	infoPtr->resultCode = Tcl_EvalObjv(infoPtr->interp, objc, objv, 0);
  
  	if (infoPtr->resultCode != TCL_OK) {
	    Tcl_AddErrorInfo(infoPtr->interp,
		    "\n    (-compare command)");
	    return order;
	}

	/*
	 * Parse the result of the command.
	 */

	if (Tcl_GetIntFromObj(infoPtr->interp,
		Tcl_GetObjResult(infoPtr->interp), &order) != TCL_OK) {
	    Tcl_ResetResult(infoPtr->interp);
	    Tcl_AppendToObj(Tcl_GetObjResult(infoPtr->interp),
		    "-compare command returned non-numeric result", -1);
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
    }
    if (!infoPtr->isIncreasing) {
	order = -order;
    }
    return order;
}

/*
 *----------------------------------------------------------------------
 *
 * DictionaryCompare
 *
 *	This function compares two strings as if they were being used in
 *	an index or card catalog.  The case of alphabetic characters is
 *	ignored, except to break ties.  Thus "B" comes before "b" but
 *	after "a".  Also, integers embedded in the strings compare in
 *	numerical order.  In other words, "x10y" comes after "x9y", not
 *      before it as it would when using strcmp().
 *
 * Results:
 *      A negative result means that the first element comes before the
 *      second, and a positive result means that the second element
 *      should come first.  A result of zero means the two elements
 *      are equal and it doesn't matter which comes first.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
DictionaryCompare(left, right)
    char *left, *right;          /* The strings to compare */
{
    Tcl_UniChar uniLeft, uniRight, uniLeftLower, uniRightLower;
    int diff, zeros;
    int secondaryDiff = 0;

    while (1) {
	if (isdigit(UCHAR(*right)) /* INTL: digit */
		&& isdigit(UCHAR(*left))) { /* INTL: digit */
	    /*
	     * There are decimal numbers embedded in the two
	     * strings.  Compare them as numbers, rather than
	     * strings.  If one number has more leading zeros than
	     * the other, the number with more leading zeros sorts
	     * later, but only as a secondary choice.
	     */

	    zeros = 0;
	    while ((*right == '0') && (isdigit(UCHAR(right[1])))) {
		right++;
		zeros--;
	    }
	    while ((*left == '0') && (isdigit(UCHAR(left[1])))) {
		left++;
		zeros++;
	    }
	    if (secondaryDiff == 0) {
		secondaryDiff = zeros;
	    }

	    /*
	     * The code below compares the numbers in the two
	     * strings without ever converting them to integers.  It
	     * does this by first comparing the lengths of the
	     * numbers and then comparing the digit values.
	     */

	    diff = 0;
	    while (1) {
		if (diff == 0) {
		    diff = UCHAR(*left) - UCHAR(*right);
		}
		right++;
		left++;
		if (!isdigit(UCHAR(*right))) { /* INTL: digit */
		    if (isdigit(UCHAR(*left))) { /* INTL: digit */
			return 1;
		    } else {
			/*
			 * The two numbers have the same length. See
			 * if their values are different.
			 */

			if (diff != 0) {
			    return diff;
			}
			break;
		    }
		} else if (!isdigit(UCHAR(*left))) { /* INTL: digit */
		    return -1;
		}
	    }
	    continue;
	}

	/*
	 * Convert character to Unicode for comparison purposes.  If either
	 * string is at the terminating null, do a byte-wise comparison and
	 * bail out immediately.
	 */

	if ((*left != '\0') && (*right != '\0')) {
	    left += Tcl_UtfToUniChar(left, &uniLeft);
	    right += Tcl_UtfToUniChar(right, &uniRight);
	    /*
	     * Convert both chars to lower for the comparison, because
	     * dictionary sorts are case insensitve.  Covert to lower, not
	     * upper, so chars between Z and a will sort before A (where most
	     * other interesting punctuations occur)
	     */
	    uniLeftLower = Tcl_UniCharToLower(uniLeft);
	    uniRightLower = Tcl_UniCharToLower(uniRight);
	} else {
	    diff = UCHAR(*left) - UCHAR(*right);
	    break;
	}

        diff = uniLeftLower - uniRightLower;
        if (diff) {
	    return diff;
	} else if (secondaryDiff == 0) {
	    if (Tcl_UniCharIsUpper(uniLeft) &&
		    Tcl_UniCharIsLower(uniRight)) {
		secondaryDiff = -1;
	    } else if (Tcl_UniCharIsUpper(uniRight)
		    && Tcl_UniCharIsLower(uniLeft)) {
		secondaryDiff = 1;
	    }
        }
    }
    if (diff == 0) {
	diff = secondaryDiff;
    }
    return diff;
}

/*
 *----------------------------------------------------------------------
 *
 * AppendObjCmd --
 *
 *	This procedure is invoked to process the "append" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result value.
 *
 * Side effects:
 *	A variable's value may be changed.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
AppendObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj *varValuePtr, *newValuePtr;
    register List *listRepPtr;
    register Tcl_Obj **elemPtrs;
    int numElems, numRequired, createdNewObj, createVar, i, j;

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "varName ?value value ...?");
	return TCL_ERROR;
    }
    if (objc == 2) {
	newValuePtr = Tcl_ObjGetVar2(interp, objv[1], (Tcl_Obj *) NULL,
		(TCL_LEAVE_ERR_MSG));
	if (newValuePtr == NULL) {
	    /*
	     * The variable doesn't exist yet. Just create it with an empty
	     * initial value.
	     */
	    
	    Tcl_Obj *nullObjPtr = Tcl_NewObj();
	    newValuePtr = Tcl_ObjSetVar2(interp, objv[1], NULL,
		    nullObjPtr, TCL_LEAVE_ERR_MSG);
	    if (newValuePtr == NULL) {
		Tcl_DecrRefCount(nullObjPtr); /* free unneeded object */
		return TCL_ERROR;
	    }
	}
    } else {
	/*
	 * We have arguments to append. We used to call Tcl_SetVar2 to
	 * append each argument one at a time to ensure that traces were run
	 * for each append step. We now append the arguments all at once
	 * because it's faster. Note that a read trace and a write trace for
	 * the variable will now each only be called once. Also, if the
	 * variable's old value is unshared we modify it directly, otherwise
	 * we create a new copy to modify: this is "copy on write".
	 */

	createdNewObj = 0;
	createVar = 1;
	varValuePtr = Tcl_ObjGetVar2(interp, objv[1], NULL, 0);
	if (varValuePtr == NULL) {
	    /*
	     * We couldn't read the old value: either the var doesn't yet
	     * exist or it's an array element. If it's new, we will try to
	     * create it with Tcl_ObjSetVar2 below.
	     */
	    
	    char *p, *varName;
	    int nameBytes, i;

	    varName = Tcl_GetStringFromObj(objv[1], &nameBytes);
	    for (i = 0, p = varName;  i < nameBytes;  i++, p++) {
		if (*p == '(') {
		    p = (varName + nameBytes-1);	
		    if (*p == ')') { /* last char is ')' => array ref */
			createVar = 0;
		    }
		    break;
		}
	    }
	    varValuePtr = Tcl_NewObj();
	    createdNewObj = 1;
	} else if (Tcl_IsShared(varValuePtr)) {	
	    varValuePtr = Tcl_DuplicateObj(varValuePtr);
	    createdNewObj = 1;
	}

	/*
	 * Convert the variable's old value to a list object if necessary.
	 */

	if (varValuePtr->typePtr != ListxListType) {
	    int result = ListxListType->setFromAnyProc(interp, varValuePtr);
	    if (result != TCL_OK) {
		if (createdNewObj) {
		    Tcl_DecrRefCount(varValuePtr); /* free unneeded obj. */
		}
		return result;
	    }
	}
	listRepPtr = (List *) varValuePtr->internalRep.otherValuePtr;
	elemPtrs = listRepPtr->elements;
	numElems = listRepPtr->elemCount;

	/*
	 * If there is no room in the current array of element pointers,
	 * allocate a new, larger array and copy the pointers to it.
	 */
	
	numRequired = numElems + (objc-2);
	if (numRequired > listRepPtr->maxElemCount) {
	    int newMax = (2 * numRequired);
	    Tcl_Obj **newElemPtrs = (Tcl_Obj **)
		ckalloc((unsigned) (newMax * sizeof(Tcl_Obj *)));
	    
	    memcpy((VOID *) newElemPtrs, (VOID *) elemPtrs,
		    (size_t) (numElems * sizeof(Tcl_Obj *)));
	    listRepPtr->maxElemCount = newMax;
	    listRepPtr->elements = newElemPtrs;
	    ckfree((char *) elemPtrs);
	    elemPtrs = newElemPtrs;
	}

	/*
	 * Insert the new elements at the end of the list.
	 */

	for (i = 2, j = numElems;  i < objc;  i++, j++) {
            elemPtrs[j] = objv[i];
            Tcl_IncrRefCount(objv[i]);
        }
	listRepPtr->elemCount = numRequired;

	/*
	 * Invalidate and free any old string representation since it no
	 * longer reflects the list's internal representation.
	 */

	Tcl_InvalidateStringRep(varValuePtr);

	/*
	 * Now store the list object back into the variable. If there is an
	 * error setting the new value, decrement its ref count if it
	 * was new and we didn't create the variable.
	 */
	
	newValuePtr = Tcl_ObjSetVar2(interp, objv[1], NULL, varValuePtr,
		TCL_LEAVE_ERR_MSG);
	if (newValuePtr == NULL) {
	    if (createdNewObj && !createVar) {
		Tcl_DecrRefCount(varValuePtr); /* free unneeded obj */
	    }
	    return TCL_ERROR;
	}
    }

    /*
     * Set the interpreter's object result to refer to the variable's value
     * object.
     */

    Tcl_SetObjResult(interp, newValuePtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ReverseObjCmd --
 *
 *	This procedure is invoked to process the "reverse" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
ReverseObjCmd(dummy, interp, objc, objv)
    ClientData dummy;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument values. */
{
    /* Syntax: reverse list
     */

    int       listc, result;
    Tcl_Obj** listv, *res;

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "list");
	return TCL_ERROR;
    }

    res = objv [1];

    /*
     * Make sure the list argument is a list object.
     */

    result = Tcl_ListObjGetElements(interp, res, &listc, &listv);

    if (result != TCL_OK) {
	return result;
    }

    /*
     * Make sure that we have an unshared list object for our work.
     */

    if (Tcl_IsShared (res)) {	
        res = Tcl_DuplicateObj (res);
    }

    result = Listx_ReverseInPlace (interp, res);

    if (result != TCL_OK) {
	return result;
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UniqueObjCmd --
 *
 *	This procedure is invoked to process the "unique" sub command
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
UniqueObjCmd(dummy, interp, objc, objv)
    ClientData dummy;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument values. */
{
    /* Syntax: unique list
     */

    int listc, result;
    Tcl_Obj **listv, *input;

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "list");
	return TCL_ERROR;
    }

    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    input  = objv [1];
    result = Tcl_ListObjGetElements(interp, input, &listc, &listv);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * All of the above was just preparation. The algorithm to remove
     * duplicates is thus:
     *
     * 1. An empty list or one with a single element cannot contain
     *	  duplicates.
     *
     * 2. For a list with more than one element add the first to the
     *    result and a hashtable. Then scan the remaining elements and
     *    look them up in the hash table. Add them to the result and
     *    the hashtable if they were not found.
     *
     *    The string representation of the list elements is used as
     *    key into the hashtable.
     *
     * This algorithm preserves the order of elements in the input.
     */

    if (listc < 2) {
        Tcl_SetObjResult (interp, input);
    } else {
        Tcl_HashEntry* hPtr;
        Tcl_HashTable  bag;
        Tcl_Obj*       res;
	int i, new;
	char* key;

	Tcl_InitHashTable (&bag, TCL_STRING_KEYS);

	res  = Tcl_NewListObj (1, listv);

	Tcl_CreateHashEntry (&bag,
	     Tcl_GetStringFromObj (listv [0], NULL),
	     &new);

	for (i=1; i < listc; i++) {
	    key  = Tcl_GetStringFromObj (listv [i], NULL);
	    hPtr = Tcl_FindHashEntry (&bag, key);

	    if (hPtr == (Tcl_HashEntry*) NULL) {
	        result = Tcl_ListObjAppendElement (interp, res, listv [i]);

		if (result != TCL_OK) {
		    Tcl_DecrRefCount (res);
		    Tcl_DeleteHashTable (&bag);
		    return result;
		}

		Tcl_CreateHashEntry (&bag, key, &new);
	    }
	}

	Tcl_DeleteHashTable (&bag);
	Tcl_SetObjResult (interp, res);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * AssignObjCmd --
 *
 *	This procedure is invoked to process the "assign" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result value.
 *
 * Side effects:
 *	Several variable's values may be changed.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
AssignObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    int listc, result, i, j, top, remains;
    Tcl_Obj **listv, *input, *res;
    Tcl_Obj* nullObj = (Tcl_Obj*) NULL;

    if (objc < 2) {
	Tcl_WrongNumArgs (interp, 2, objv-1, "list ?varName varName...?");
	return TCL_ERROR;
    }

    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    input  = objv [1];
    result = Tcl_ListObjGetElements (interp, input, &listc, &listv);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * list            = X0...Xn-1    (n = listc)
     * obj [2..objc-1] = V0...Vk-1    (k = objc-2)
     *
     * -------------------------------------------------------
     *
     * Sideeffects:
     *		Vi = Xi	for i = 0 ... min (n,k) - 1 = top - 1
     *		Vj = {}	for j = n ... k - 1, if n < k
     *
     * Result:
     *		{}		for n <= k
     *		{Xk..Xn-1}	for k < n
     */

    top = listc < (objc-2) ? listc : objc-2;

    for (i=0, j=2; i < top; i++, j++) {
        res = Tcl_ObjSetVar2 (interp, objv [j], (Tcl_Obj*) NULL,
		      listv [i], TCL_PARSE_PART1 | TCL_LEAVE_ERR_MSG);

	if (res == (Tcl_Obj*) NULL)
	    goto handle_error;
    }

    if (listc < (objc-2)) {
        nullObj = Tcl_NewObj ();
	Tcl_IncrRefCount (nullObj);

        for (; j < objc; j++) {
	    res = Tcl_ObjSetVar2 (interp, objv [j], (Tcl_Obj*) NULL,
			nullObj, TCL_PARSE_PART1 | TCL_LEAVE_ERR_MSG);

	    if (res == (Tcl_Obj*) NULL)
	        goto handle_error;
	}

	Tcl_DecrRefCount (nullObj);
    }

    remains = listc - (objc-2);

    if (remains > 0) {
        Tcl_SetObjResult (interp, Tcl_NewListObj (remains, listv + (objc-2)));
    }

    return TCL_OK;

handle_error:
    if (nullObj != (Tcl_Obj*) NULL) {
        Tcl_DecrRefCount (nullObj);
    }
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * ConcatObjCmd --
 *
 *	This procedure is invoked to process the "concat" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result value.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
ConcatObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj* res, **listv;
    int i, listc, result;

    if (objc < 2) {
	Tcl_WrongNumArgs (interp, 2, objv-1, "list...");
	return TCL_ERROR;
    }

    res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

    for (i=1; i < objc; i++) {
        result = Tcl_ListObjGetElements (interp, objv [i], &listc, &listv);
	if (result != TCL_OK) {
	    goto handle_error;
	}

	result = Tcl_ListObjAppendList (interp, res, objv [i]);

	if (result != TCL_OK) {
	    goto handle_error;
	}
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;

handle_error:
    Tcl_DecrRefCount (res);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * RemoveObjCmd --
 *
 *	This procedure is invoked to process the "remove" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A new Tcl list object formed by removing zero or more elements of
 *	a list.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
RemoveObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj* res, **listv;
    int listc, result;

    if (objc < 2) {
	Tcl_WrongNumArgs (interp, 2, objv-1, "list ?index...?");
	return TCL_ERROR;
    }

    result = Tcl_ListObjGetElements(interp, objv [1], &listc, &listv);

    if (result != TCL_OK) {
	return result;
    }

    if (objc == 2) {
      /* No selecting objects, just the list. The result is a full
       * duplicate. Avoid copying, just generate an additional
       * reference to the input.
       */

      Tcl_SetObjResult (interp, objv [1]);
      return TCL_OK;

    } else if (objc == 3) {
      /* A single selecting object. The result is specified by that
       * object.
       */

      result = Listx_RemoveSFromListByObj (interp, objv [1], objv [2], &res);

    } else {
      /* More than one selector. Take all selections and make them
       * the final result.
       */

      result = Listx_RemoveMFromListByObj (interp, objv [1],
					   objc-2, (Tcl_Obj**) objv+2,
					   &res);
    }

    if (result != TCL_OK) {
      return result;
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UnknownObjCmd --
 *
 *	This procedure is invoked to process all unknown sub commands
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
UnknownObjCmd(dummy, interp, objc, objv)
    ClientData dummy;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument values. */
{
    /* Call '::listx::unknown' to handle unknown commands at the script
     * level.
     */

    if (ListxUnknownCmdObj == (Tcl_Obj*) NULL) {
        /* Retain the error message from the caller */
        return TCL_ERROR;
    } else {
        Tcl_Obj** newObjv;
	int i, res;

	Tcl_ResetResult (interp);

	newObjv = (Tcl_Obj**) Tcl_Alloc (sizeof (Tcl_Obj*) * objc);

	if (newObjv == (Tcl_Obj**) NULL) {
	    return TCL_ERROR;
	}

	newObjv [0] = ListxUnknownCmdObj;
	Tcl_IncrRefCount (newObjv [0]);

	for (i=1; i < objc; i++) {
	  newObjv [i] = objv [i];
	  Tcl_IncrRefCount (newObjv [i]);
	}

	/*
	 * Execution in the scope of our caller.
	 * Not in the global scope.
	 */

	res = Tcl_EvalObjv (interp, objc, newObjv, 0);
			 
	for (i=0; i < objc; i++) {
	    Tcl_DecrRefCount (newObjv [i]);
	}
	Tcl_Free ((char*) newObjv);

	return res;
      }
}

/*
 *----------------------------------------------------------------------
 *
 * MapObjCmd --
 *
 *	This procedure is invoked to process the "map" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
MapObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    register Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj* res, **listv;
    int listc, result;

    if (objc != 3) {
	Tcl_WrongNumArgs (interp, 2, objv-1, "list command");
	return TCL_ERROR;
    }

    result = Tcl_ListObjGetElements(interp, objv [1], &listc, &listv);

    if (result != TCL_OK) {
	return result;
    }

    result = Listx_Map (interp, objv [1], ListxScriptMapProc,
			(ClientData) objv [2], &res);

    if (result != TCL_OK) {
	return result;
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ListxScriptMapProc --
 *
 *	Special map procedure, lifts mapping into the script level.
 *	
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
ListxScriptMapProc (clientData, interp, inObj, resultObj)
     ClientData  clientData;
     Tcl_Interp* interp;
     Tcl_Obj*    inObj;
     Tcl_Obj**   resultObj;
{
#if USE_EVAL_OBJV
    Tcl_Obj* cmdv [2];
    Tcl_Obj* cmdObj = (Tcl_Obj*) clientData;
    int      result;

    cmdv [0] = cmdObj; Tcl_IncrRefCount (cmdObj);
    cmdv [1] = inObj;  Tcl_IncrRefCount (inObj);

    result = Tcl_EvalObjv (interp, 2, cmdv, 0);

    Tcl_DecrRefCount (cmdObj);
    Tcl_DecrRefCount (inObj);

    if (result != TCL_OK) {
        return result;
    }

    *resultObj = Tcl_GetObjResult (interp);
    Tcl_IncrRefCount (*resultObj);
    Tcl_ResetResult (interp);
#else
    Tcl_Obj* cmdv [2], *cmd;
    Tcl_Obj* cmdObj = (Tcl_Obj*) clientData;
    int      result;

    cmdv [0] = inObj;
    cmdv [1] = Tcl_NewListObj (1, cmdv);
    cmdv [0] = cmdObj;

    cmd = Tcl_ConcatObj (2, cmdv);
    Tcl_IncrRefCount (cmd);

    result = Tcl_EvalObj (interp, cmd);

    Tcl_DecrRefCount (cmd);
    cmd = (Tcl_Obj*) NULL;
    Tcl_DecrRefCount (cmdv [1]);

    if (result != TCL_OK) {
      return result;
    }

    *resultObj = Tcl_GetObjResult (interp);
    Tcl_IncrRefCount (*resultObj);
    Tcl_ResetResult (interp);
#endif

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * MergeObjCmd --
 *
 *	This procedure is invoked to process the "merge" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
MergeObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    register Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj* res;
    int result, i, groups;

    static char *options [] = {
        "-group", "--",
	NULL
    };
    enum options {
        SPLIT_GROUP, SPLIT_LAST
    };

    groups = 0;

    for (i = 1; i < objc; i++) {
	char* name;
	int   index;

	name = Tcl_GetString(objv[i]);
	if (name[0] != '-') {
	    break;
	}
	if (Tcl_GetIndexFromObj(interp, objv [i], options, "option", TCL_EXACT,
		&index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch ((enum options) index) {
	case SPLIT_GROUP:
	    groups = 1;
	    break;

	case SPLIT_LAST:
	    i++;
	    goto endOfForLoop;
	}
    }

    endOfForLoop:
    if (objc - i < 1) {
	Tcl_WrongNumArgs (interp, 2, objv-1, "?-group? ?--? list...");
	return TCL_ERROR;
    }

    objc -= i;
    objv += i;

    /* End of option processing.
     * ----------------------------------- */

    result = Listx_Merge (interp, objc, objv, groups, &res);

    if (result != TCL_OK) {
	return result;
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * EmptyObjCmd --
 *
 *	This procedure is invoked to process the "empty" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
EmptyObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    register Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj *res;
    int result, empty;

    if (objc != 2) {
	Tcl_WrongNumArgs (interp, 2, objv-1, "list");
	return TCL_ERROR;
    }

    result = Listx_Empty (interp, objv [1], &empty);

    if (result != TCL_OK) {
	return result;
    }

    res = Tcl_NewBooleanObj (empty);

    if (res == (Tcl_Obj*) NULL) {
        return TCL_ERROR;
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FlattenObjCmd --
 *
 *	This procedure is invoked to process the "flatten" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
FlattenObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    register Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj *res;
    int result, full, i;

    static char *options [] = {
        "-full", "--",
	NULL
    };
    enum options {
        FLATTEN_FULL, FLATTEN_LAST
    };

    full = 0;

    for (i = 1; i < objc; i++) {
	char* name;
	int   index;

	name = Tcl_GetString(objv[i]);
	if (name[0] != '-') {
	    break;
	}
	if (Tcl_GetIndexFromObj(interp, objv [i], options, "option", TCL_EXACT,
		&index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch ((enum options) index) {
	case FLATTEN_FULL:
	    full = 1;
	    break;

	case FLATTEN_LAST:
	    i++;
	    goto endOfForLoop;
	}
    }

    endOfForLoop:
    if (objc - i < 1) {
	Tcl_WrongNumArgs(interp, 2, objv-1, "?-full? ?--? list");
	return TCL_ERROR;
    }

    objc -= i;
    objv += i;

    result = Listx_Flatten (interp, objv [0], full, &res);

    if (result != TCL_OK) {
	return result;
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FoldObjCmd --
 *
 *	This procedure is invoked to process the "fold" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
FoldObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    register Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Tcl_Obj* res, **listv;
    int listc, result;

    if (objc != 4) {
	Tcl_WrongNumArgs (interp, 2, objv-1, "list initialvalue command");
	return TCL_ERROR;
    }

    result = Tcl_ListObjGetElements(interp, objv [1], &listc, &listv);

    if (result != TCL_OK) {
	return result;
    }

    result = Listx_Fold (interp, objv [1], objv [2], ListxScriptFoldProc,
			(ClientData) objv [3], &res);

    if (result != TCL_OK) {
	return result;
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ListxScriptFoldProc --
 *
 *	Special fold procedure, lifts folding into the script level.
 *	
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
ListxScriptFoldProc (clientData, interp, accumulatorObj, inObj, resultObj)
     ClientData  clientData;
     Tcl_Interp* interp;
     Tcl_Obj*    accumulatorObj;
     Tcl_Obj*    inObj;
     Tcl_Obj**   resultObj;
{
#if USE_EVALOBJV
    Tcl_Obj* cmdv [3];
    Tcl_Obj* cmdObj = (Tcl_Obj*) clientData;
    int      result;

    cmdv [0] = cmdObj;          Tcl_IncrRefCount (cmdObj);
    cmdv [1] = accumulatorObj;  Tcl_IncrRefCount (accumulatorObj);
    cmdv [2] = inObj;           Tcl_IncrRefCount (inObj);

    result = Tcl_EvalObjv (interp, 3, cmdv, 0);

    Tcl_DecrRefCount (cmdObj);
    Tcl_DecrRefCount (accumulatorObj);
    Tcl_DecrRefCount (inObj);

    if (result != TCL_OK) {
        return result;
    }

    *resultObj = Tcl_GetObjResult (interp);
    Tcl_IncrRefCount (*resultObj);
    Tcl_ResetResult (interp);
#else
    Tcl_Obj* cmdv [3];
    Tcl_Obj* cmd;
    Tcl_Obj* cmdObj = (Tcl_Obj*) clientData;
    int      result;

    cmdv [0] = accumulatorObj;
    cmdv [1] = Tcl_NewListObj (1, cmdv);
    cmdv [0] = inObj;
    cmdv [2] = Tcl_NewListObj (1, cmdv);
    cmdv [0] = cmdObj;

    cmd = Tcl_ConcatObj (3, cmdv);
    Tcl_IncrRefCount (cmd);

    result = Tcl_EvalObj (interp, cmd);

    Tcl_DecrRefCount (cmd);
    cmd = (Tcl_Obj*) NULL;
    Tcl_DecrRefCount (cmdv [1]);
    Tcl_DecrRefCount (cmdv [2]);

    if (result != TCL_OK) {
        return result;
    }

    *resultObj = Tcl_GetObjResult (interp);
    Tcl_IncrRefCount (*resultObj);
    Tcl_ResetResult (interp);
#endif

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SplitObjCmd --
 *
 *	This procedure is invoked to process the "split" sub command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
SplitObjCmd(dummy, interp, objc, objv)
    ClientData dummy;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument values. */
{
    int       listc, stride, result, groups, i;
    Tcl_Obj **listv, *res;

    static char *options [] = {
        "-group", "--",
	NULL
    };
    enum options {
        SPLIT_GROUP, SPLIT_LAST
    };

    groups = 0;

    for (i = 1; i < objc; i++) {
	char* name;
	int   index;

	name = Tcl_GetString(objv[i]);
	if (name[0] != '-') {
	    break;
	}
	if (Tcl_GetIndexFromObj(interp, objv [i], options, "option", TCL_EXACT,
		&index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch ((enum options) index) {
	case SPLIT_GROUP:
	    groups = 1;
	    break;

	case SPLIT_LAST:
	    i++;
	    goto endOfForLoop;
	}
    }

    endOfForLoop:
    if (objc - i < 1) {
	Tcl_WrongNumArgs (interp, 2, objv-1, "?-group? ?--? list stride");
	return TCL_ERROR;
    }

    /* End of option processing.
     * ----------------------------------- */

    result = Tcl_ListObjGetElements(interp, objv [objc-2], &listc, &listv);

    if (result != TCL_OK) {
	return result;
    }

    result = Tcl_GetIntFromObj(interp, objv [objc-1], &stride);

    if (result != TCL_OK) {
	return result;
    }

    result = Listx_Split (interp, objv [objc-2], groups, stride, &res);

    if (result != TCL_OK) {
	return result;
    }

    Tcl_SetObjResult (interp, res);
    return TCL_OK;
}

