# listx.decls -- -*- tcl -*-
#
#	This file contains the declarations for all supported public
#	functions that are exported by the Listx library via the stubs table.
#	This file is used to generate the listxDecls.h file.
#	

library listx

# Define the list interface with several sub interfaces:
#     listxInt	 - generic private

interface listx
hooks {listxInt}

# Declare each of the functions in the public Listx interface.  Note that
# every index should never be reused for a different function in order
# to preserve backwards compatibility.

#
# Grouping of functionality in the interface.
#
# 1) Functions dealing with selectors (creation, usage)
# 2) Map & Fold
# 3) Split & Merge
# 4) Other
#
# The groups are separated by at least one reserved slot to allow for
# future expansion.
#

#
# 1 -  Selecting information.
#  a - Creating a selector from an object (possibly changing its type)
#  b - Creating selectors from searching through lists.
#  c - Using a selector to get at the specified elements of a list
#  d - Using a selector to get at a list *not* containing the specified
#      elements
#

# Creates a new selector object from the information in the structure.

declare 0 generic {
    Tcl_Obj* Listx_NewSelectorObj (listSelector* CONST ls)
}

# Fills the selector with the information from the 'obj'ect, possibly
# changing its type. Returns TCL_ERROR upon failure, leaves
# an error message in the 'interp' in that case too. No copying, we get
# a pointer to the internal representation.

declare 1 generic {
    int Listx_GetSelectorFromObj (Tcl_Interp* interp, Tcl_Obj* obj,
	listSelector** ls)
}

# Fills the selector with the index (or all indices) matching the pattern.
# The search starts at the element with the given 'offset'.

declare 2 generic {
    int Listx_Search (Tcl_Interp* interp, Tcl_Obj* listObj,
	Tcl_Obj* patternObj, long int offset, lSearchModes mode,
	int inverted, int nocase, int all, listSelector* ls)
}

# Use a selector to create an object containing just the specified
# elements of the original list. Two variants are available, one
# directly using the selector structure, the other taking a selector
# object. Two more variants allow the selection by multiple selectors.
# In-place variants are not yet defined. The problem is to define what
# the result should be for LIST_SEL_SINGLE. After solving that these
# variants go into the reserved slots.

declare 3 generic {
    int Listx_SelectSFromList (Tcl_Interp* interp, Tcl_Obj* listObj,
    	listSelector* CONST ls, Tcl_Obj** resultObj)
}

declare 4 generic {
    int Listx_SelectSFromListByObj (Tcl_Interp* interp, Tcl_Obj* listObj,
    	Tcl_Obj* selectorObj, Tcl_Obj** resultObj)
}

declare 5 generic {
    int Listx_SelectMFromList (Tcl_Interp* interp, Tcl_Obj* listObj,
    	int lsc, listSelector** CONST lsv, Tcl_Obj** resultObj)
}

declare 6 generic {
    int Listx_SelectMFromListByObj (Tcl_Interp* interp, Tcl_Obj* listObj,
    	int soc, Tcl_Obj** sov, Tcl_Obj** resultObj)
}

# Use a selector to create an object *not* containing the specified
# elements of the original list. Two variants are available, one
# directly using the selector structure, the other taking a selector
# object.

declare 7 generic {
    int Listx_RemoveSFromList (Tcl_Interp* interp, Tcl_Obj* listObj,
    	listSelector* CONST ls, Tcl_Obj** resultObj)
}

declare 8 generic {
    int Listx_RemoveSFromListByObj (Tcl_Interp* interp, Tcl_Obj* listObj,
    	Tcl_Obj* selectorObj, Tcl_Obj** resultObj)
}

declare 9 generic {
    int Listx_RemoveMFromList (Tcl_Interp* interp, Tcl_Obj* listObj,
    	int lsc, listSelector** CONST lsv, Tcl_Obj** resultObj)
}

declare 10 generic {
    int Listx_RemoveMFromListByObj (Tcl_Interp* interp, Tcl_Obj* listObj,
    	int soc, Tcl_Obj** sov, Tcl_Obj** resultObj)
}

# Reserved for future use
#  declare 11 generic {
#  }
#  declare 12 generic {
#  }
#  declare 13 generic {
#  }
#  declare 14 generic {
#  }
#  declare 15 generic {
#  }
#  declare 16 generic {
#  }
#  declare 17 generic {
#  }
#  declare 18 generic {
#  }
#  declare 19 generic {
#  }

#
# 2 - Map & Fold. Applying commands to lists and their elements.
#

# Mapping objects to other objects by way of a function.
# Folding a sequence of objects into a single one.

declare 20 generic {
    int Listx_Map (Tcl_Interp* interp, Tcl_Obj* listObj,
	Listx_MapProc* mapProc, ClientData clientData,
	Tcl_Obj** resultObj)
}

declare 21 generic {
    int Listx_Fold (Tcl_Interp* interp, Tcl_Obj* listObj, Tcl_Obj* initialObj,
	Listx_FoldProc* foldProc, ClientData clientData,
	Tcl_Obj** resultObj)
}

# Reserved for future use
#  declare 22 generic {
#  }
#  declare 23 generic {
#  }
#  declare 24 generic {
#  }
#  declare 25 generic {
#  }
#  declare 26 generic {
#  }
#  declare 27 generic {
#  }
#  declare 28 generic {
#  }
#  declare 29 generic {
#  }

#
# 3 - Split & Merge.
#

declare 30 generic {
    int Listx_Split (Tcl_Interp* interp, Tcl_Obj* listObj, int group,
	int stride, Tcl_Obj** resultObj)
}

declare 31 generic {
    int Listx_Merge (Tcl_Interp* interp, int listc, Tcl_Obj* CONST* listv,
    	int group, Tcl_Obj** resultObj)
}

# Reserved for future use
#  declare 32 generic {
#  }
#  declare 33 generic {
#  }
#  declare 34 generic {
#  }
#  declare 35 generic {
#  }
#  declare 36 generic {
#  }
#  declare 37 generic {
#  }
#  declare 38 generic {
#  }
#  declare 39 generic {
#  }

#
# 4 - Other functionality, misc.
#

declare 40 generic {
    int Listx_Empty (Tcl_Interp* interp, Tcl_Obj* listObj, int* emptyPtr)
}

# Reversing the order of elements in a list (Variants for both copying and
# in-place manipulation are available).

declare 41 generic {
    int Listx_Reverse (Tcl_Interp* interp, Tcl_Obj* listObj,
	Tcl_Obj** resultObj)
}

declare 42 generic {
    int Listx_ReverseInPlace (Tcl_Interp* interp, Tcl_Obj* listObj)
}

declare 43 generic {
    int Listx_Flatten (Tcl_Interp* interp, Tcl_Obj* listObj, int full,
	Tcl_Obj** resultObj)
}

#
# listx.decls -- -*- tcl -*-

# Reserved for future use
#  declare 44 generic {
#  }
