/* 
 * remove.c --
 *
 *	This file contains the C-APIs for removing objects by index.
 *
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: remove.c,v 1.2 2000/01/14 23:44:16 aku Exp $
 */

#include "listxInt.h"

/*
 *----------------------------------------------------------------------
 *
 * Listx_RemoveSFromList --
 *
 *	Takes the elements whose indices are NOT specified by the selector
 *	from the list object and packs them into a new list object (of
 *	refcount 0).
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	A new list object in 'resultObj' if there were no errors.
 *
 *----------------------------------------------------------------------
 */

int
Listx_RemoveSFromList (interp, listObj, ls, resultObj)
     Tcl_Interp*         interp;
     Tcl_Obj*            listObj;
     listSelector* CONST ls;
     Tcl_Obj**           resultObj;
{
  /*
   * Make sure the list argument is a list object and get its length and
   * a pointer to its array of element pointers.
   */

  int      result, listc, i, index, start, end, stride;
  int*     flags;
  Tcl_Obj* res, **listv;

  result = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);

  if (result != TCL_OK) {
    return TCL_ERROR;
  }

  /* We use an array to hold the markers showing which objects to place
   * into the the result (flag [i] == 0) and which not (flag [i] == 1).
   * The code to fill the array depends on the type of the selector. The
   * code to actually generate the doesn't.
   */

  flags = (int*) Tcl_Alloc (listc * sizeof (int));
  memset (flags, '\0', listc * sizeof (int));

  switch (ls->type) {
  case LIST_SEL_SINGLE:
    result = ListxConvertToFullIndex (&ls->info.index, listc, &index);
    if (result == TCL_OK) {
      flags [index] = 1;
    }
    break;

  case LIST_SEL_MULTIPLE:
    for (i=0; i < ls->info.multiple.num; i++) {
      result = ListxConvertToFullIndex (&ls->info.multiple.index [i], listc,
					&index);

      if (result == TCL_OK) {
	flags [index] = 1;
      }
    }
    break;

  case LIST_SEL_RANGE:
    ListxConvertToFullIndex (&ls->info.range.start, listc, &start);
    ListxConvertToFullIndex (&ls->info.range.end,   listc, &end);
    stride = ls->info.range.stride;

    if ((start <= listc) && (end >= 0) && (stride <= listc)) {
      if (start < 0) {
	start = (stride - ((-start) % stride)) % stride;
      }
      if (end >= listc) {
	end = listc-1;
      }

      for (i=start; i <= end; i += stride) {
	flags [i] = 1;
      }
    }
    break;
  }

  /* Ok, the flags are set, now use them to get the correct elements.
   */

  res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

  for (i=0; i < listc; i++) {
    if (!flags [i]) {
	result = Tcl_ListObjAppendElement (interp, res, listv [i]);

	if (result != TCL_OK) {
	  Tcl_DecrRefCount (res);
	  Tcl_Free ((char*) flags);
	  return TCL_ERROR;
	}
    }
  }

  Tcl_Free ((char*) flags);

  *resultObj = res;
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Listx_RemoveSFromListByObj --
 *
 *	Takes the elements whose indices are NOT specified by the selector
 *	object from the list object and packs them into a new list object
 *	(of refcount 0).
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	A new list object in 'resultObj' if there were no errors.
 *
 *----------------------------------------------------------------------
 */

int
Listx_RemoveSFromListByObj (interp, listObj, selectorObj, resultObj)
     Tcl_Interp* interp;
     Tcl_Obj*    listObj;
     Tcl_Obj*    selectorObj;
     Tcl_Obj**   resultObj;
{
  int           result, idx;
  listSelector* ls, ls_static;

  /* Special case code to avoid shimmering of 'int' objects.
   * Future: Use the 'Integer interface' if it is provided.
   *
   * In the special case a separate 'static' list selector is filled
   * with the relevant information and then given to the low level
   * code.
   */

  if (ListxIntType == selectorObj->typePtr) {
    ls = &ls_static;

    ls->type = LIST_SEL_SINGLE;
    result = Tcl_GetIntFromObj (interp, selectorObj, &idx);

    if (result == TCL_OK) {
      ls->info.index.location   = idx;
      ls->info.index.offsetFrom = LOFF_START;
    }
  } else {
    result = Listx_GetSelectorFromObj (interp, selectorObj, &ls);
  }

  if (result != TCL_OK) {
    return result;
  }

  return Listx_RemoveSFromList (interp, listObj, ls, resultObj);
}

/*
 *----------------------------------------------------------------------
 *
 * Listx_RemoveMFromList --
 *
 *	Takes for each of the specified selector structures the elements
 *	not selected by it, packs them into a new list and then makes these
 *	lists item in the final result, again a new list.
 *
 * Results:
 *	A standard tcl result code
 *
 * Side effects:
 *	resultObj is set to a new list object containing the lists
 *	containing the not-selected elements. The value is not changed if
 *	there were problems.
 *
 *----------------------------------------------------------------------
 */

int
Listx_RemoveMFromList (interp, listObj, lsc, lsv, resultObj)
     Tcl_Interp*          interp;
     Tcl_Obj*             listObj;
     int                  lsc;
     listSelector** CONST lsv;
     Tcl_Obj**            resultObj;
{
  int          result, i;
  Tcl_Obj*     res;
  Tcl_Obj*     item;
  listSelector ls;

  /* Special check: If all selectors are of type ..._SINGLE we have to
   * handle them as a single selector of type ..._MULTIPLE!
   */

  for (i=0; i < lsc; i++) {
    if (lsv [i]->type != LIST_SEL_SINGLE) {
      goto normal; /* Different types, skip to standard behaviour */
    }
  }

  ls.type = LIST_SEL_MULTIPLE;
  ls.info.multiple.num   = lsc;
  ls.info.multiple.index = (lIndex*) Tcl_Alloc (lsc * sizeof (lIndex));

  for (i=0; i < lsc; i++) {
    memcpy (&ls.info.multiple.index [i],
	    &lsv [i]->info.index,
	    sizeof (lIndex));
  }

  result = Listx_RemoveSFromList (interp, listObj, &ls, &res);

  Tcl_Free ((char*) ls.info.multiple.index);

  if (result != TCL_OK) {
    return TCL_ERROR;
  }

  goto done;

normal:

  res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

  for (i=0; i < lsc; i++) {
    result = Listx_RemoveSFromList (interp, listObj, lsv [i], &item);

    if (result != TCL_OK) {
      Tcl_DecrRefCount (res);
      return TCL_ERROR;
    }

    result = Tcl_ListObjAppendElement (interp, res, item);

    if (result != TCL_OK) {
      Tcl_DecrRefCount (res);
      return TCL_ERROR;
    }
  }

done:
  *resultObj = res;
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Listx_RemoveMFromListByObj --
 *
 *	Takes for each of the specified selector objects the elements
 *	not selected by it, packs them into a new list and then makes these
 *	lists item in the final result, again a new list.
 *
 * Results:
 *	A standard tcl result code
 *
 * Side effects:
 *	resultObj is set to a new list object containing the lists
 *	containing the not-selected elements. The value is not changed if
 *	there were problems.
 *
 *----------------------------------------------------------------------
 */

int
Listx_RemoveMFromListByObj (interp, listObj, soc, sov, resultObj)
     Tcl_Interp* interp;
     Tcl_Obj*    listObj;
     int         soc;
     Tcl_Obj**   sov;
     Tcl_Obj**   resultObj;
{
  int            result, i, idx;
  listSelector** ls, *ls_int;

  ls     = (listSelector**) Tcl_Alloc (soc * sizeof (listSelector*));
  ls_int = (listSelector*)  Tcl_Alloc (soc * sizeof (listSelector));

  for (i=0; i< soc; i++) {
    /* Special case code to avoid shimmering of 'int' objects.
     * Future: Use the 'Integer interface' if it is provided.
     * See [x] too.
     */

    if (ListxIntType == sov [i]->typePtr) {

      ls [i] = &ls_int [i];
      ls [i]->type = LIST_SEL_SINGLE;

      result = Tcl_GetIntFromObj (interp, sov [i], &idx);

      if (result == TCL_OK) {
	ls [i]->info.index.location   = idx;
	ls [i]->info.index.offsetFrom = LOFF_START;
      }
    } else {
      result = Listx_GetSelectorFromObj (interp, sov [i], &(ls [i]));
    }

    if (result != TCL_OK) {
      Tcl_Free ((char*) ls);
      Tcl_Free ((char*) ls_int);
      return result;
    }
  }

  result = Listx_RemoveMFromList (interp, listObj, soc, ls, resultObj);

  Tcl_Free ((char*) ls);
  Tcl_Free ((char*) ls_int);
  return result;
}

