/* 
 * mapfold.c --
 *
 *	This file contains the C-APIs for mapping and folding objects.
 *
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: split.c,v 1.2 2000/01/14 23:44:16 aku Exp $
 */

#include "listxInt.h"

/*
 *----------------------------------------------------------------------
 *
 * Listx_Split --
 *
 *	Splits the provided list into a list of lists.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	See above.
 *
 *----------------------------------------------------------------------
 */

int
Listx_Split (interp, listObj, group, stride, resultObj)
     Tcl_Interp* interp;
     Tcl_Obj*    listObj;
     int         group;
     int         stride;
     Tcl_Obj**   resultObj;
{
    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    Tcl_Obj** listv, *res, *sub, *item;
    int       listc, result, i, j, k, nsub, nelem;

    result = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * How many new lists do we get in the result ?
     * And how many elements do they have ?
     *
     * Example input: 	{a b c d e f g} 2
     *
     * group = 0  =>   {a c e g} {b d f {}}
     *                  stride                             sublists
     *			listc/stride + ((listc%stride)!=0) elements per sublist
     *
     * group = 1  =>   {a b} {c d} {e f} {g {}}
     *                  stride                             elements per sublist
     *			listc/stride + ((listc%stride)!=0) sublists
     */

    res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

    if (res == (Tcl_Obj*) NULL) {
        return TCL_ERROR;
    }

    if (group) {
        nsub  = listc / stride + (listc % stride ? 1 : 0);
	nelem = stride;

	for (i=0, k=0; i < nsub; i++) {
	    sub = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

	    if (sub == (Tcl_Obj*) NULL) {
	        Tcl_DecrRefCount (res);
	        return TCL_ERROR;
	    }

	    for (j=0; j < nelem; j++, k++) {
	        if (k >= listc) {
		    item = Tcl_NewObj ();

		    if (item == (Tcl_Obj*) NULL) {
		        Tcl_DecrRefCount (sub);
		        Tcl_DecrRefCount (res);
			return TCL_ERROR;
		    }
		} else {
		    item = listv [k];
		}

		result = Tcl_ListObjAppendElement (interp, sub, item);

		if (result != TCL_OK) {
		    Tcl_DecrRefCount (sub);
		    Tcl_DecrRefCount (res);
		    return result;
		}
	    }

	    result = Tcl_ListObjAppendElement (interp, res, sub);

	    if (result != TCL_OK) {
	        Tcl_DecrRefCount (res);
		return result;
	    }
	}
    } else {
        nsub  = stride;
	nelem = listc / stride + (listc % stride ? 1 : 0);

	for (i=0; i < nsub; i++) {
	    sub = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

	    if (sub == (Tcl_Obj*) NULL) {
	        Tcl_DecrRefCount (res);
	        return TCL_ERROR;
	    }

	    for (j=0, k=i; j < nelem; j++, k += nsub) {
	        if (k >= listc) {
		    item = Tcl_NewObj ();

		    if (item == (Tcl_Obj*) NULL) {
		        Tcl_DecrRefCount (sub);
		        Tcl_DecrRefCount (res);
			return TCL_ERROR;
		    }
		} else {
		    item = listv [k];
		}

		result = Tcl_ListObjAppendElement (interp, sub, item);

		if (result != TCL_OK) {
		    Tcl_DecrRefCount (sub);
		    Tcl_DecrRefCount (res);
		    return result;
		}
	    }

	    result = Tcl_ListObjAppendElement (interp, res, sub);

	    if (result != TCL_OK) {
	        Tcl_DecrRefCount (res);
		return result;
	    }
	}
    }

    *resultObj = res;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Listx_Merge --
 *
 *	Merges the provided lists into a single list. The complementary
 *	function to Listx_Split.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	See above.
 *
 *----------------------------------------------------------------------
 */

int 
Listx_Merge (interp, listc, listv, group, resultObj)
     Tcl_Interp*     interp;
     int             listc;
     Tcl_Obj* CONST* listv;
     int             group;
     Tcl_Obj**       resultObj;
{
    /*
     * Make sure that all objects in 'listv' are lists. Remember
     * their sizes and the poiinters to their element arrays.
     */
    /*
    Tcl_Obj** listv, *res, *sub, *item;
    int       listc, result, i, j, k, nsub, nelem;
    */

    int       result, i, j, stride;
    int*       objc = (int*)       NULL;
    Tcl_Obj*** objv = (Tcl_Obj***) NULL;
    Tcl_Obj*   res  = (Tcl_Obj*)   NULL;

    objc = (int*)      Tcl_Alloc (listc * sizeof (int));

    if (objc == (int*) NULL) {
        goto error;
    }

    objv = (Tcl_Obj***) Tcl_Alloc (listc * sizeof (Tcl_Obj**));

    if (objv == (Tcl_Obj***) NULL) {
        goto error;
    }

    /*
     * How do we merge the inputs together.
     *
     * 1) group == 0: {a c e g} {b d f {}}      =>  {a b c d e f g {}}
     * 2) group == 1: {a b} {c d} {e f} {g {}}  =>  {a b c d e f g {}}
     *
     * Always stride == Max {Length (listv[i]) | 0 <= i < listc }
     */

    stride = 0;
    for (i=0; i < listc; i++) {
      result = Tcl_ListObjGetElements (interp, listv [i],
				       &(objc[i]), &(objv[i]));
      if (result != TCL_OK) {
	goto error;
      }

      if (objc [i] > stride) {
	stride = objc [i];
      }
    }

    res = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

    if (res == (Tcl_Obj*) NULL) {
	goto error;
    }

    if (group) {
        /*
	 * The simple case: concatenate the lists together and use
	 * empty elements to fill the gaps (after the lists shorter
	 * than stride elements).
	 *
	 * 2) group == 1: {a b} {c d} {e f} {g {}}  =>  {a b c d e f g {}}
	 */

        Tcl_Obj** sub;
        Tcl_Obj*  item;

        for (i=0; i < listc; i++) {
	    sub = objv [i];

	    for (j=0; j < objc [i]; j++) {
	        result = Tcl_ListObjAppendElement (interp, res, sub [j]);

		if (result != TCL_OK) {
		    goto error;
		}
	    }

	    for ( ; j < stride; j++) {
	        item = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

		if (item == (Tcl_Obj*) NULL) {
		    goto error;
		}

	        result = Tcl_ListObjAppendElement (interp, res, item);

		if (result != TCL_OK) {
		    goto error;
		}
	    }
	}

    } else {
        /*
	 * The more complex case: go through the lists in parallel and add
	 * the found items to the result. More complex because we have to
	 * check that the incoming lists are long enough.
	 *
	 * 1) group == 0: {a c e g} {b d f {}}      =>  {a b c d e f g {}}
	 */

        Tcl_Obj** sub;
        Tcl_Obj*  item;

        for (i=0; i < stride; i++) {

	    for (j=0; j < listc; j++) {
	        sub = objv [j];

	        if (i < objc [j]) {
		    item = sub [i];
		} else {
		    item = Tcl_NewListObj (0, (Tcl_Obj**) NULL);

		    if (item == (Tcl_Obj*) NULL) {
		        goto error;
		    }
		}

	        result = Tcl_ListObjAppendElement (interp, res, item);

		if (result != TCL_OK) {
		    goto error;
		}
	    }
	}
    }

    *resultObj = res;
    return TCL_OK;

error:
    /* Cleanup memory before returning with an error.
     */

    if (objc != (int*) NULL) {
        Tcl_Free ((char*) objc);
    }
    if (objv != (Tcl_Obj***) NULL) {
        Tcl_Free ((char*) objv);
    }
    if (res != (Tcl_Obj*) NULL) {
	Tcl_DecrRefCount (res);
    }
    return TCL_ERROR;
}
